////////// GenoMus v.0.01
////////// http://www.lopezmontes.es/genomus.html



outlets = 5;

//////////////////////
// global variables
var maxLevels = 5; // maximal number of parenthesis levels
var isBranch; // indentify if the actual expression is a subexpression
var maxNumberFuncions = 200; // maximal number of functions per expresssion
var openParenth = 0; // current level of parenthesis 
var argumentStatus = []; // 0 = writing first argument, 1 = writing second argument, etc;
var currentFunctionName; // current function being written
var newExpression; // new expression or branch created by core functions createExpression
var growingExpression; // provisional evolved function from currentExpression; 
var provisoryExpression = [];
var usedFunctionsArray = []; // array with used functions to control the
var currentExpression; // last expression (functions of functions created)
var currentExpressionRendered = []; // current expression with primitives rendered 
var savedExpressions = []; // array of manually saved expressions
var generatedScoreLength;
var goodScore = 1; // control if the expression is well-formed
var minScoreSize = 1; // min number of events in the generated scores
var maxScoreSize = 500; // max number of events in the generated scores
var goodScoreSize = 0; // boolean for testing the good size
var maxSearchIterarions = 100; // max number of iteraration searching a score
var readablePrinting = 1; // boolean for activate a tabulated print, best for human reading
var dispersionAlea = 97940987898; // (0, 4) -> like logistic equation. For uniform-like distribution, uses very long numbers
var chromaticQuantiz = 12; // number of divisions per octave (12 for usual chromatic division of the keyboard)

var alwaysAutoreference; // always include autoreferences in new branchs
var verboseOutput = 0;


/////////////////////////////
// library of functions used to construct the expression





///////////////////// library of functions for creation of new functions
// goal: try to minimize calculus

/////////// array of every functions with its associated output. At the same time create the links functionName->number, to expression encoding
// link a function name with a number in both directions
var function2index = [];
var index2function = [];

var functionsType = [];
function addFunctionsTypeItem(functionName, outputType) {
    functionsType[functionName] = outputType;
    function2index[functionName] = index2function.length;
    index2function[index2function.length] = functionName;
} 


var intervalMatrixGlobal = [3,6,4,7,8,4,1,4,7,4,2,1]; // global vertical intervals matrix 
function setIntervalMatrixGlobal (score) {
    intervalMatrixGlobal = score;
}
addFunctionsTypeItem("intervalMatrixGlobal", "score"); 


function initIntervalMatrixGlobal (score) {
    intervalMatrixGlobal = matrixUnidim(12, 1, 12);
    post("intervalMatrixGlobal = " + intervalMatrixGlobal + "\n");
}


// index primitive functions of main script
addFunctionsTypeItem("rInt", "integ"); 
addFunctionsTypeItem("rFlo", "float"); 
addFunctionsTypeItem("intPrim", "integ"); 
addFunctionsTypeItem("floPrim", "float"); 

// arithmetic functions

function sqr(value) { 
    return value*value; 
}
addFunctionsTypeItem("sqr", "float"); // index this function with its output type

function sum(value_1, value_2) { 
    return value_1 + value_2; 
}
addFunctionsTypeItem("sum", "float"); 

function dif(value_1, value_2) { 
    return value_1 - value_2; 
}
addFunctionsTypeItem("dif", "float"); 

function absDif(value_1, value_2) { 
    return Math.abs(value_1 - value_2); 
}
addFunctionsTypeItem("absDif", "float"); 

function prod(value_1, value_2) { 
    return value_1 * value_2; 
}
addFunctionsTypeItem("prod", "float"); 

function ratio(value_1, value_2) { 
    if ( value_2 == 0 ) { 
        goodScore = 0; 
        return [-1];
    }
    return value_1 / value_2; 
}
addFunctionsTypeItem("ratio", "float"); 

function mod(value_1, value_2) {
    return value_1 % value_2;
}
addFunctionsTypeItem("mod", "float"); 

function invers(value) { 
    if ( value == 0 ) { 
        goodScore = 0; 
        return [-1];
    }
    return 1 / value; 
}
addFunctionsTypeItem("invers", "float"); 

/* 
function pow(value_1, value_2) { 
    var power = Math.pow(value_1, value_2);
    if ( Math.abs(power) > 100000000 ) {
        goodScore = 0; 
        post("INFINF");
        return [0]
    }
    return power; 
}
addFunctionsTypeItem("pow", "float");
*/ 

// return value_1th root of value_2
function root(value_1, value_2) { 
    result = Math.pow(Math.abs(value_2), 1/value_1);
    if (Math.abs(result) > 1000000000 ) { 
        post("too big number");
        return [-1];
    } 
    return result; 
}
addFunctionsTypeItem("root", "float"); 

function pyth(value_1, value_2) { 
    return root(sum(sqr(value_1),sqr(value_2)),2); 
}
addFunctionsTypeItem("pyth", "float"); 

function log(value_1, value_2) { 
    if ( value_1 < 0 || value_2 <= 0 || value_2 == 1 ) { 
        goodScore = 0; 
        return [-1];
    }
    return Math.log(value_1) / Math.log(value_2); 
}
addFunctionsTypeItem("log", "float"); 

function sin(value) {
    return Math.sin(value);
}
addFunctionsTypeItem("sin", "float"); 

function scoSin (score, value) {
    for (n=0; n<score.length; n++) {
        score[n] = Math.abs(Math.sin(score[n]) * value);
    }
    return score;
}
addFunctionsTypeItem("scoSin", "score"); 

function cos(value) {
    return Math.cos(value);
}
addFunctionsTypeItem("cos", "float"); 

function scoCos (score, value) {
    for (n=0; n<score.length; n++) {
        score[n] = Math.abs(Math.cos(score[n]) * value);
    }
    return score;
}
addFunctionsTypeItem("scoCos", "score");

function asin(value) {
    return Math.asin(value % 1);
}
addFunctionsTypeItem("asin", "float"); 

function acos(value) {
    return Math.acos(value % 1);
}
addFunctionsTypeItem("acos", "float"); 

function tan(value) {
    return Math.tan(value % 1.57079); // avoid results too long
}
addFunctionsTypeItem("tan", "float"); 

function atan(value) {
    return Math.atan(value);
}
addFunctionsTypeItem("atan", "float"); 

function abs(value) {
    return Math.abs(value);
}
addFunctionsTypeItem("abs", "float"); 

// return a random float from the interval [value_1,value_2)
function randFlo(value_1, value_2) {
    return randomSeed() * (value_2 - value_1) + value_1;
}
addFunctionsTypeItem("randFlo", "float"); 

// return a random int from the interval [value_1,value_2] (both included)
function randInt(value_1, value_2) {
    return Math.round(randomSeed() * (Math.round(value_2) - Math.round(value_1)) + Math.round(value_1));
}
addFunctionsTypeItem("randInt", "integ"); 


//////////// functions with musical matrix
/// musical matrix (or score) is normalized according to a score type previously defined

// return score % value
function scoMod(score, value) {
    if ( value == 0 ) { 
        goodScore = 0; 
        return [-1];
    }	
    for (a=0; a<score.length; a++) {
        score[a] = score[a] % value;
    }
    return score;
}
addFunctionsTypeItem("scoMod", "score"); 

// transpose a score value interval
function scoTransp(score, value) {
    for (i=0; i<score.length; i++) {
        score[i] = score[i] + value;
    }
    return score;
}
addFunctionsTypeItem("scoTransp", "score"); 

// composes a score transposing score_1 to each score_2 value
function scoTranspMatrix(score_1, score_2) {
    if ( score_1.length * score_2.length > maxScoreSize ) { 
        goodScore = 0; 
        return [-1];
    }
    var composedScore = [];
    for ( a=0; a<score_2.length; a++ ) {
        composedScore = composedScore.concat(scoTransp(score_1, score_2[a]));
    }
    return composedScore;
} 
addFunctionsTypeItem("scoTranspMatrix", "score"); 

// repite a musical matrix integ times
function scoRep(score, integ) {
    if (score.length * integ <= maxScoreSize) {
        var tempScore = [];
        for (i=0; i<integ; i++) {
            tempScore = tempScore.concat(score);
        }
        return tempScore;
    }
    else { 
        goodScore = 0; 
        return [-1];
    }
}
addFunctionsTypeItem("scoRep", "score"); 

// concatenate two musical matrix
function scoConcat(score_1, score_2) {
    if (score_1.length + score_2.length <= maxScoreSize) {
        return score_1.concat(score_2);
    }
    else {
        goodScore = 0; 
        return [-1];
    }
}
addFunctionsTypeItem("scoConcat", "score"); 

// return retrogradation of score  
function scoRetrog(score) {
    return score.reverse();
}
addFunctionsTypeItem("scoRetrog", "score"); 

// return invertion of score 
function scoInvert(score) {
    var scoreCopy = score.slice(0); // copies array without loops
    for( ele=1; ele<score.length; ele++ ) {
        score[ele] = score[ele-1] - (scoreCopy[ele] - scoreCopy[ele-1]);
    }
    return score;
}
addFunctionsTypeItem("scoInvert", "score"); 

// return score expanded according to factor value_1, using first as axis 
function scoExpand(score, value_1) {
    var expandedScore = score.slice(0); 
    for( ele=1; ele<score.length; ele++ ) {
        expandedScore[ele] = score[0] + (expandedScore[ele] - score[0]) * value_1;
    }
    return expandedScore;
}
addFunctionsTypeItem("scoExpand", "score"); 

// iterated version of scoExpandIter, expanding score value_3 times, from factor value_1 to value_2
function scoExpandIter(score, value_1, value_2, value_3) {
    if ( score.length * Math.abs(value_3) > maxScoreSize ) { 
        goodScore = 0; 
        return [-1];
    }
    var expansionIndex = scoSteps(value_1, value_2, value_3);
    var expandedScore = [];
    for ( a=0; a<Math.round(Math.abs(value_3)); a++ ) {
        expandedScore = expandedScore.concat(scoExpand(score, expansionIndex[a]));
    }
    return expandedScore;
}
addFunctionsTypeItem("scoExpandIter", "score"); 

// return a new score extract with value_2 notes starting in value_1 note. If value_2 is negative, excerpt is returned retrogradated
function scoExcerpt(score, value_1, value_2) {
    value_1 = Math.round(Math.abs(value_1)) % score.length;
    if ( value_2 == 0 ) { // if value_2 is 0, excerpt have only 1 value
        value_2 = 1; 
    }
    var step = 1;
    if ( value_2 < 0 ) { step = -1 };
    var excerpt = [];
    var newEle = 0;
    for ( ele = value_1; ele != value_1 + value_2 && ele < score.length && ele >= 0; ele = ele + step ) {
        excerpt[newEle] = score[ele];
        newEle++;
    }
    return excerpt;
}
addFunctionsTypeItem("scoExcerpt", "score"); 

// return a series of excerpts from score_1, from initial note to score_2 notes
function scoExcerptMulti(score_1, score_2) {
    if ( absSumArray(score_2) > maxScoreSize ) {
        goodScore = 0; 
        return [-1];        
    }
    var composedScore = [];
    var sco1length = score_1.length;
    for(it=0; it<score_2.length; it++) {
        composedScore = composedScore.concat(scoExcerpt(score_1, 0, score_2[it] % sco1length));
    }
    return composedScore;
}
addFunctionsTypeItem("scoExcerptMulti", "score"); 

// return a series of excerpts from score_1, from score_2 notes to score_3 notes
function scoExcerptMulti2(score_1, score_2, score_3) {
    if ( absSumArray(score_2) > maxScoreSize || absSumArray(score_3) > maxScoreSize ) {
        goodScore = 0; 
        return [-1];        
    }
    var composedScore = [];
    var sco1length = score_1.length;
    for(it=0; it<Math.min(score_2.length, score_3.length); it++) {
        composedScore = composedScore.concat(scoExcerpt(score_1, score_2[it] % sco1length, score_3[it] % sco1length));
    }
    return composedScore;
}
addFunctionsTypeItem("scoExcerptMulti2", "score"); 

// return a series of excerpts from score_1, from score_2 notes to score_3 notes, transposed score_4 intervals
function scoExcerptMultiTransp(score_1, score_2, score_3, score_4) {
    if ( absSumArray(score_2) > maxScoreSize || absSumArray(score_3) > maxScoreSize ) {
        goodScore = 0; 
        return [-1];        
    }
    var composedScore = [];
    var sco1length = score_1.length;
    for(it=0; it<Math.min(score_2.length, score_3.length, score_4.length); it++) {
        composedScore = composedScore.concat(scoTransp(scoExcerpt(score_1, score_2[it] % sco1length, score_3[it] % sco1length),score_4[it]));
    }
    return composedScore;
}
addFunctionsTypeItem("scoExcerptMultiTransp", "score"); 

// return a new score with the same notes in a new random order
function scoPermut(score) {
    var scoreCopy = score.slice(0);
    var permutScore = [];
    var selectedElem;
    do {
        selectedElem = parseInt( scoreCopy.length * randomSeed() );
        permutScore.push( scoreCopy[selectedElem] );
        scoreCopy.splice(selectedElem, 1);
    } while ( scoreCopy.length > 0 );
    return permutScore;	
}
addFunctionsTypeItem("scoPermut", "score"); 

// iterated version of scoPermut: return a new score with the same notes in a new random order, value_1 times
function scoPermutIter(score, value_1) {
    if ( score.length * Math.abs(value_1) > maxScoreSize ) { 
        goodScore = 0; 
        return [-1];
    }
    var scoreCopy;
    var permutScore = [];
    var selectedElem;
    for ( a=0; a<Math.round(Math.abs(value_1)); a++ ) {
        scoreCopy = score.slice(0);
        do {
            selectedElem = parseInt( scoreCopy.length * randomSeed() );
            permutScore.push( scoreCopy[selectedElem] );
            scoreCopy.splice(selectedElem, 1);
        } while ( scoreCopy.length > 0 );
    }
    return permutScore;	
}
addFunctionsTypeItem("scoPermutIter", "score"); 

// return array from value_1 to value_2 in value_3 steps
function scoSteps(value_1, value_2, value_3) {
    value_3 = Math.round(Math.abs(value_3));
    if ( value_3 < 2 && value_3 > -2 ) { 
        goodScore = 0; 
        return [-1];
    }    
    var stepsArray = [];
    var step = (value_2 - value_1) / (value_3 - 1);
    for ( a=0; a<value_3; a++ ) {
        stepsArray[a] = step * a + value_1;
    }
    return stepsArray;
}
addFunctionsTypeItem("scoSteps", "score"); 

// create a random integer score, with integ_1 items, between boundaries integ_2 and integ_3)
function scoRand(integ_1, integ_2, integ_3) {
    if (integ_1 > maxScoreSize) {
        goodScore = 0; 
        return [-1];
    }
    var tempScore = [];
    for (i=0; i<integ_1; i++) {
        tempScore[i] = rInt(integ_2, integ_3);
    }   
    return tempScore;
} 
addFunctionsTypeItem("scoRand", "score"); 

// create a random array (without limitations os scoreSize), with integ_1 items, between boundaries integ_2 and integ_3)
function matrixUnidim(integ_1, integ_2, integ_3) {
    var tempScore = [];
    for (i=0; i<integ_1; i++) {
        tempScore[i] = rInt(integ_2, integ_3);
    }   
    return tempScore;
} 
addFunctionsTypeItem("matrixUnidim", "score"); 

// create a random float score, with integ_1 items, between boundaries float_1 and float_2)
function scoRandFlo(integ_1, float_1, float_2) {
    if (integ_1 > maxScoreSize) {
        goodScore = 0; 
        return [-1];
    }
    var tempScore = [];
    for (i=0; i<integ_1; i++) {
        tempScore[i] = rFlo(float_1, float_2);
    }   
    return tempScore;
} 
addFunctionsTypeItem("scoRandFlo", "score");

// create a primitive score from scratch     
function scoRandPrim() { 
    return eval(scoRand(rInt(3, 15), 21, 108));
}   
addFunctionsTypeItem("scoRandPrim", "score"); 

// return an array of integers from integ_1 to integ_2 
function scoEnumInt(integ_1, integ_2) {
    if (Math.abs(integ_1 - integ_2) > maxScoreSize) {
         goodScore = 0; 
        return [-1];
    }       
    var tempArray = [];
    if (integ_1>integ_2) { var step = -1 }
    else { var step = 1 }
    for (i=integ_1; i!=integ_2+step; i=i+step) {
        tempArray[tempArray.length] = i;
    }
    return tempArray;
}
addFunctionsTypeItem("scoEnumInt", "score"); 

// return the score with some differences according to float_1 (0=no mutation, 0.999=max mutation), and value_2 (range of mutation)
function scoMutate(score, float_1, value_2) {
    float_1 = float_1 % 1;
    for(i=0; i<score.length; i++) {
        if (randomSeed() < float_1) {
            score[i] = score[i] + rFlo(value_2*(-1), value_2);
        }
    }
    return score;
}
addFunctionsTypeItem("scoMutate", "score"); 

function scoRender(score) {
    return score;
}
addFunctionsTypeItem("scoRender", "score"); 

// generates a random pitch class set
function scoPCGen(integ) {
    // number of elements of PCS from 1 to chromaticQuantiz
    var numberPitchClasses = (Math.abs(integ) - 1) % chromaticQuantiz + 1;
    var pitchClassSet = [];
    var newElem;
    do {
        newElem = rInt(0,chromaticQuantiz-1);
        if ( pitchClassSet.indexOf(newElem) == -1 ) {
            pitchClassSet.push(newElem);
        }
    } while ( pitchClassSet.length < numberPitchClasses );
    return pitchClassSet.sort(function(a,b){return a-b});
}  
addFunctionsTypeItem("scoPCGen", "score"); 

function scoPCGenPrim() {
    return scoPCGen(rInt(1,11));
}
addFunctionsTypeItem("scoPCGenPrim", "score"); 

// adjusts the notes of a score to a pitch class set
function scoPitchClass(score, pcset) {
    //var pcset = pcsFromScore(score_2);
    var octaveNotes;
    var tempScore = [];
    // this loop look for the more accurate conversion to a note from the pcset
    for (b=0; b<score.length; b++) {
        var c = 0;
        do {
            c++;
            octaveNotes = Math.floor(score[b]/chromaticQuantiz)*chromaticQuantiz;
            tempScore[b] = pcset[c-1] + octaveNotes;
        } while ( ( Math.abs(score[b] - (octaveNotes + pcset[c-1]) ) > Math.abs(score[b] - (octaveNotes + pcset[c]) ) ) && c < pcset.length )
    }
    return tempScore; 
}
addFunctionsTypeItem("scoPitchClass", "score"); 

/////// BABBAGE functions

function scoNameBABBAGE () {
    return [58,57,58,58,57,55,52];
}
addFunctionsTypeItem("scoNameBABBAGE", "score"); 

function scoNameBABBAGEintervals () {
    return [-1,1,0,-1,-2,-3];
}
addFunctionsTypeItem("scoNameBABBAGEintervals", "score");

function scoNameADA () {
    return [57,50,57];
}
addFunctionsTypeItem("scoNameADA", "score"); 

function scoNameADAintervals () {
    return [-7,7];
}
addFunctionsTypeItem("scoNameADAintervals", "score");

// returns score with value new notes, calculated according to a 2-dim. matrix of the last two intervals
function scoADA(score, value) {
    /*if ( value + score.length >= score.length ) { 
        goodScore = 0;
        return [-1];
    } */
    var newScore = score.slice(0);
    // // 7x7 table to calculate new interval (interval from -3 to 3)
    var biggestDescendingInterval = -7;
    var biggestAscendingInterval = 7; 
    var rangeOfIntervalsInMatrix = 15;
    var lastInterval, prelastInterval;
    var nextIntervalMatrix = matrixUnidim(225, biggestDescendingInterval, biggestAscendingInterval); // table to calculate new interval
    var lastInterval =  Math.round((newScore[newScore.length - 1] - newScore[newScore.length - 2]) % 7);
    var prelastInterval = Math.round((newScore[newScore.length - 2] - newScore[newScore.length - 3]) % 7);
    for ( newEl = 0; newEl < Math.abs(value); newEl++ ) {
        lastInterval = Math.round((newScore[newScore.length - 1] - newScore[newScore.length - 2]) % 7);
        prelastInterval = Math.round((newScore[newScore.length - 2] - newScore[newScore.length - 3]) % 7);
        newInterval = nextIntervalMatrix[(prelastInterval+7)*15 + lastInterval+7];
        if ( newInterval == undefined ) { newInterval = 0 };
        newScore.push( newScore[newScore.length - 1] + newInterval );
    }
    return newScore;
}
addFunctionsTypeItem("scoADA", "score"); 

// return a second voice, with another matrix with vertical intervals
function scoADAnewVoice(score) {
    /*if ( value + score.length >= score.length ) { 
        goodScore = 0;
        return [-1];
    } */
    var newInterval;
    var lowScore = score.slice(0);
    var newVoiceScore = [];
    newVoiceScore.push(lowScore[0], lowScore[1])
    // 7x7 table to calculate new interval (interval from -3 to 3)
    var biggestDescendingInterval = 0;
    var biggestAscendingInterval = 14; 
    var rangeOfIntervalsInMatrix = 15;
    var lastInterval, prelastInterval;
    var nextIntervalMatrix = matrixUnidim(225, biggestDescendingInterval, biggestAscendingInterval); // table to calculate new interval
    post("MATRIX = " + nextIntervalMatrix + "\n");
    var lastInterval = Math.round((lowScore[lowScore.length - 1] - lowScore[lowScore.length - 2]) % 7);
    var prelastInterval = Math.round((lowScore[lowScore.length - 2] - lowScore[lowScore.length - 3]) % 7);
    for ( newEl = 2; newEl < score.length; newEl++ ) {
        lastInterval = Math.round((lowScore[newEl] - lowScore[newEl - 1]) % 7);
        prelastInterval = Math.round((lowScore[newEl - 1] - lowScore[newEl - 2]) % 7);
        newInterval = nextIntervalMatrix[(prelastInterval+7)*15 + lastInterval+7];
        if ( newInterval == undefined ) { newInterval = 0 };
        newVoiceScore.push(lowScore[newEl] + newInterval);
    }
    return newVoiceScore;
}
addFunctionsTypeItem("scoADAnewVoice", "score"); 

// return a second voice, with another matrix with vertical intervals
function scoBABBAGEnewVoice(score) {
    /*if ( value + score.length >= score.length ) { 
        goodScore = 0;
        return [-1];
    } */
    var newInterval;
    var lowScore = score.slice(0);
    var newVoiceScore = [];
    newVoiceScore.push(lowScore[0], lowScore[1])
    // 7x7 table to calculate new interval (interval from -3 to 3)
    var biggestDescendingInterval = 0;
    var biggestAscendingInterval = 20; 
    var rangeOfIntervalsInMatrix = 21;
    var lastInterval, prelastInterval;
    var nextIntervalMatrix = matrixUnidim(441, biggestDescendingInterval, biggestAscendingInterval); // table to calculate new interval
    post("MATRIX = " + nextIntervalMatrix + "\n");
    var lastInterval = Math.round((lowScore[lowScore.length - 1] - lowScore[lowScore.length - 2]) % 12);
    var prelastInterval = Math.round((lowScore[lowScore.length - 2] - lowScore[lowScore.length - 3]) % 12);
    for ( newEl = 2; newEl < score.length; newEl++ ) {
        lastInterval = Math.round((lowScore[newEl] - lowScore[newEl - 1]) % 12);
        prelastInterval = Math.round((lowScore[newEl - 1] - lowScore[newEl - 2]) % 12);
        newInterval = nextIntervalMatrix[(prelastInterval+10)*21 + lastInterval+10];
        if ( newInterval == undefined ) { newInterval = 7 };
        newVoiceScore.push(lowScore[newEl] + newInterval);
    }
    return newVoiceScore;
}
addFunctionsTypeItem("scoBABBAGEnewVoice", "score"); 

// scans 2 scores, and return a new 3rd voice, considering only vertical intervals between 2 initial voices to determine third voice
function scoBABBAGEharmonicVoice(score_1, score_2) {
    var newScore = [];
    var biggestDescendingInterval = -15;
    var biggestAscendingInterval = 15; 
    var nextIntervalMatrix = matrixUnidim(12, biggestDescendingInterval, biggestAscendingInterval); // table to calculate new interval
    post("MATRIX = " + nextIntervalMatrix + "\n");
    for (d=0; d<Math.min(score_1.length, score_2.length); d++) {
        newScore[d] = Math.max(score_1[d], score_2[d]) + nextIntervalMatrix[Math.round(Math.abs(score_1[d] - score_2[d])) % 12];
    }
    return newScore;
}
addFunctionsTypeItem("scoBABBAGEharmonicVoice", "score"); 


// identical as precedent function, but using a global vertical intervals array
function scoBABBAGEharmonicVoiceGlobal(score_1, score_2) {
    initIntervalMatrixGlobal();
    var newScore = [];
    post("MATRIX = " + intervalMatrixGlobal + "\n");
    for (d=0; d<Math.min(score_1.length, score_2.length); d++) {
        newScore[d] = Math.max(score_1[d], score_2[d]) + intervalMatrixGlobal[Math.round(Math.abs(score_1[d] - score_2[d])) % 12];
    }
    return newScore;
}
addFunctionsTypeItem("scoBABBAGEharmonicVoiceGlobal", "score"); 

// return a score alternating notes from score_1 y score_2 (both of identical length)
function scoJoin2Scores(score_1, score_2) {
    var joinScoreLength = Math.min(score_1.length, score_2.length);
    if (joinScoreLength > maxScoreSize) {
        goodScore = 0;
        return [-1];
    }
    var composedScore = [];
    for (i=0; i<joinScoreLength; i++) {
        composedScore.push(score_1[i], score_2[i]);
    }
    return composedScore;
}
addFunctionsTypeItem("scoJoin2Scores", "score"); 

// return a score alternating notes from score_1, score_2 and score_3 (all of identical length)
function scoJoin3Scores(score_1, score_2, score_3) {
    var joinScoreLength = Math.min(score_1.length, score_2.length, score_3.length);
    if (joinScoreLength > maxScoreSize) {
        goodScore = 0;
        return [-1];
    }
    var composedScore = [];
    for (i=0; i<joinScoreLength; i++) {
        composedScore.push(score_1[i], score_2[i], score_3[i]);
    }
    return composedScore;
}
addFunctionsTypeItem("scoJoin3Scores", "score");

// return a score alternating notes from score_1, score_2 and score_3 (all of identical length)
function scoJoin4Scores(score_1, score_2, score_3, score_4) {
    var joinScoreLength = Math.min(score_1.length, score_2.length, score_3.length, score_4.length);
    if (joinScoreLength > maxScoreSize) {
        goodScore = 0;
        return [-1];
    }
    var composedScore = [];
    for (i=0; i<joinScoreLength; i++) {
        composedScore.push(score_1[i], score_2[i], score_3[i], score_4[i]);
    }
    return composedScore;
}
addFunctionsTypeItem("scoJoin4Scores", "score");

// return a score alternating notes from score_1, score_2 and score_3 (all of identical length)
function scoJoin5Scores(score_1, score_2, score_3, score_4, score_5) {
    var joinScoreLength = Math.min(score_1.length, score_2.length, score_3.length, score_4.length, score_5.length);
    if (joinScoreLength > maxScoreSize) {
        goodScore = 0;
        return [-1];
    }
    var composedScore = [];
    for (i=0; i<joinScoreLength; i++) {
        composedScore.push(score_1[i], score_2[i], score_3[i], score_4[i], score_5[i]);
    }
    return composedScore;
}
addFunctionsTypeItem("scoJoin5Scores", "score");


// ADA 2 voices generator
function scoADA2voices(score, value) {
    var lowVoice = scoADA(score, value);
    var upperVoice = scoADAnewVoice(lowVoice);
    return scoJoin2Scores(lowVoice, upperVoice);
}
addFunctionsTypeItem("scoADA2voices", "score"); 

// ADA 3 voices generator
function scoADA3voices(score, value) {
    var lowVoice = scoADA(score, value);
    var interVoice = scoADAnewVoice(lowVoice);
    var upperVoice = scoADAnewVoice(interVoice);
    return scoJoin3Scores(lowVoice, interVoice, upperVoice);
}
addFunctionsTypeItem("scoADA3voices", "score"); 

// BABBAGE 2 voices generator
function scoBABBAGE2voices(score, value) {
    var lowVoice = scoBABBAGE(score, value);
    var upperVoice = scoBABBAGEnewVoice(lowVoice);
    return scoJoin2Scores(lowVoice, upperVoice);
}
addFunctionsTypeItem("scoBABBAGE2voices", "score"); 

// BABBAGE 3 voices generator
function scoBABBAGE3voices(score, value) {
    var lowVoice = scoBABBAGE(score, value);
    var interVoice = scoBABBAGEnewVoice(lowVoice);
    var upperVoice = scoBABBAGEnewVoice(interVoice);
    return scoJoin3Scores(lowVoice, interVoice, upperVoice);
}
addFunctionsTypeItem("scoBABBAGE3voices", "score"); 

// BABBAGE 2 voices + harmonic voice generator
function scoBABBAGE3voicesHarmonic(score, value) {
    var lowVoice = scoBABBAGE(score, value);
    var interVoice = scoBABBAGEnewVoice(lowVoice);
    var upperVoice = scoBABBAGEharmonicVoice(lowVoice, interVoice);
    return scoJoin3Scores(lowVoice, interVoice, upperVoice);
}
addFunctionsTypeItem("scoBABBAGE3voicesHarmonic", "score"); 

// BABBAGE 2 voices + 2 harmonic voices generator
function scoBABBAGE4voicesHarmonic(score, value) {
    var lowVoice = scoBABBAGE(score, value);
    var interVoice = scoBABBAGEnewVoice(lowVoice);
    var upperVoice = scoBABBAGEharmonicVoice(lowVoice, interVoice);
    var upperUpperVoice = scoBABBAGEharmonicVoice(interVoice, upperVoice);
    return scoJoin4Scores(lowVoice, interVoice, upperVoice, upperUpperVoice);
}
addFunctionsTypeItem("scoBABBAGE4voicesHarmonic", "score");

// BABBAGE 2 voices + 3 harmonic voices generator
function scoBABBAGE5voicesHarmonic(score, value) {
    var lowVoice = scoBABBAGE(score, value);
    var interVoice = scoBABBAGEnewVoice(lowVoice);
    var upperVoice = scoBABBAGEharmonicVoice(lowVoice, interVoice);
    var upperUpperVoice = scoBABBAGEharmonicVoice(interVoice, upperVoice);
    var topVoice = scoBABBAGEharmonicVoice(upperVoice, upperUpperVoice);
    return scoJoin5Scores(lowVoice, interVoice, upperVoice, upperUpperVoice, topVoice);
}
addFunctionsTypeItem("scoBABBAGE5voicesHarmonic", "score");


// BABBAGE 2 voices + 3 harmonic voices generator using identical matrix interval (as global matrix)
function scoBABBAGE5voicesHarmonicGlobal(score, value) {
    initIntervalMatrixGlobal();
    var lowVoice = scoBABBAGE(score, value);
    var interVoice = scoBABBAGEnewVoice(lowVoice);
    //intervalMatrixGlobal = matrixUnidim(12, 1, 16);
    var upperVoice = scoBABBAGEharmonicVoiceGlobal(lowVoice, interVoice);
    var upperUpperVoice = scoBABBAGEharmonicVoiceGlobal(interVoice, upperVoice);
    var topVoice = scoBABBAGEharmonicVoiceGlobal(upperVoice, upperUpperVoice);
    return scoJoin5Scores(lowVoice, interVoice, upperVoice, upperUpperVoice, topVoice);
}
addFunctionsTypeItem("scoBABBAGE5voicesHarmonicGlobal", "score");

// add a second voice following a rule based in precedents intervals
function addBABBAGEvoice(score) {
    var newVoice = scoBABBAGEnewVoice(score);
    return scoJoin2Scores(score, newVoice);
}
addFunctionsTypeItem("addBABBAGEvoice", "score"); 

// add a second voice following a rule based in precedents intervals
function addADAvoice(score) {
    var newVoice = scoADAnewVoice(score);
    return scoJoin2Scores(score, newVoice);
}
addFunctionsTypeItem("addADAvoice", "score"); 

// returns score with value new notes, calculated according to a 2-dim. matrix of the last two intervals
function scoBABBAGE(score, value) {
    /*if ( value + score.length >= score.length ) { 
        goodScore = 0;
        return [-1];
    } */ 
    var newScore = score.slice(0);   
    var biggestDescendingInterval = -11;
    var biggestAscendingInterval = 11; 
    var rangeOfIntervalsInMatrix = 23;
    var lastInterval, prelastInterval;
    var nextIntervalMatrixBABBAGE = matrixUnidim(529, biggestDescendingInterval, biggestAscendingInterval); // table to calculate new interval
    post("BABBAGE = " + nextIntervalMatrixBABBAGE + "\n");
    // introduce the constants for BABBAGE motif
    nextIntervalMatrixBABBAGE[( -1 +11)*23 + 1 +11] = 0; // BAB -> B
    nextIntervalMatrixBABBAGE[( 1 +11)*23 + 0 +11] = -1; // ABB -> A
    nextIntervalMatrixBABBAGE[( 0 +11)*23 + -1 +11] = -2; // BBA -> G
    nextIntervalMatrixBABBAGE[( -1 +11)*23 + -2 +11] = -3; // BAG -> E

    // introduce the constants for BABBAGE inverted motif    
    nextIntervalMatrixBABBAGE[( 1 +11)*23 + -1 +11] = 0; // invert(BAB -> B)
    nextIntervalMatrixBABBAGE[( -1 +11)*23 + 0 +11] = 1; // invert(ABB -> A)
    nextIntervalMatrixBABBAGE[( 0 +11)*23 + 1 +11] = 2; // invert(BBA -> G)
    nextIntervalMatrixBABBAGE[( 1 +11)*23 + 2 +11] = 3; // invert(BAG -> E)  
    
    // ADA loop: DAD -> A  
    nextIntervalMatrixBABBAGE[( -7 +11)*23 + 7 +11] = -7; // ADA -> D 
    nextIntervalMatrixBABBAGE[( 7 +11)*23 + -7 +11] = 7; // DAD -> A 
    
    
    var lastInterval = Math.round((newScore[newScore.length - 1] - newScore[newScore.length - 2]) % 12);
    var prelastInterval = Math.round((newScore[newScore.length - 2] - newScore[newScore.length - 3]) % 12);
    for ( newEl = 0; newEl < Math.abs(value); newEl++ ) {
        lastInterval = Math.round((newScore[newScore.length - 1] - newScore[newScore.length - 2]) % 12);
        prelastInterval = Math.round((newScore[newScore.length - 2] - newScore[newScore.length - 3]) % 12);
        newInterval = nextIntervalMatrixBABBAGE[(prelastInterval+11)*23 + lastInterval+11] + mod(newScore[newEl-1], 3);
        if ( isNaN(newInterval) == true ) { newInterval = 0 };
        newScore.push( newScore[newScore.length - 1] + newInterval );
    }
    return newScore;
}
addFunctionsTypeItem("scoBABBAGE", "score"); 

// return the value-th last interval from score
function intrv(score, value) {
    if ( value >= score.length ) { 
        goodScore = 0;
        return [-1];
    }
    return score[score.length - value] - score[score.length - value - 1];
}
addFunctionsTypeItem("intrv", "value"); 

// return the value-th last note from score
function nthLastNote(score, value) {
    if ( value >= score.length ) { 
        goodScore = 0;
        return [-1];
    }
    return score[score.length - value];
}
addFunctionsTypeItem("nthLastNote", "value"); 

// return a recursive score with value notes, using recursive encoded expression recur with score as initial conditions
function scoRecursio(score, recur, value) {
    if ( value + score.length >= maxScoreSize ) { 
        goodScore = 0;
        return [-1];
    }  
    //post("RECIBI recur = " + recur + "\n");
    var recursio = decodeExpr(recur);
    var recursioPrinted = expandExpr( "encodeExpr(\"" + recursio + "\")" );
    outlet(4, recursioPrinted);
    //post("RECURSIO es " + recursio + "\n");
    var s = score.slice(0);
    //post("SCORE es " + s + "\n");
    for ( newEl = 0; newEl < Math.abs(value); newEl++ ) {
        //var("nuevo elem = " + eval(recursio) + "\n");
        s.push( s[s.length - 1] + eval(recursio) );
    }
    return s;
}
function preScoRecursio(value) {
}
addFunctionsTypeItem("scoRecursio", "score");
addFunctionsTypeItem("preScoRecursio", "score");

// return a recursive score with value notes, using recursive encoded expression recur with score as initial conditions
function scoRecursioBABBAGE(recur) {
    post("RECUR recibido es " + recur + "\n");
    var recursio = decodeExpr(recur);
    var recursioPrinted = expandExpr( "encodeExpr(\"" + recursio + "\")" );
    outlet(4, recursioPrinted);
    var s = MIDI2genomusPitch([58,57,58,58,57,54,51]);
    for ( newEl = 0; newEl < 200; newEl++ ) {
        s.push( s[s.length - 1] + eval(recursio) );
    }
    post("NO CONVERT = " + s + "\n");
    post("MOD = " + scoMod(s,1) + "\n");
    post("CONVERT = " + genomusPitch2MIDI(scoMod(s,1)) + "\n");
    return genomusPitch2MIDI(scoMod(s,1));
}
function preScoRecursioBABBAGE(value) {
}
addFunctionsTypeItem("scoRecursioBABBAGE", "score");
addFunctionsTypeItem("preScoRecursioBABBAGE", "score");

//mask functions for sustitute with recursions
function preEncRecursio() {
}
function preEncRecursioBABBAGE() {
}
addFunctionsTypeItem("preEncRecursio", "prov");
addFunctionsTypeItem("preEncRecursioBABBAGE", "prov");


// reuses a subExpression from the currentExpression
function scoInter(internalExpressionAddress) {
    if ( insco_primitive.length > 0 ) {
        return eval( extractSubExpression( growingExpression, internalExpressionAddress) );
    }
    else { 
        goodScore = 0; 
        return [-1];
    }
}
function preScoInter() {
}
addFunctionsTypeItem("scoInter", "score");
addFunctionsTypeItem("preScoInter", "prov"); 

// eval value times expre, used as a literal expression (return different results in every eval)
function scoRepExpr(internalExpressionAddress, numberIter) {
    if ( insco_primitive.length == 0 ) {
        goodScore = 0; 
        return [-1];
    }
    expri_2 = (numberIter % 10) + 1;
    expressionEvaluated = extractSubExpression( growingExpression, internalExpressionAddress );
    if ( eval(expressionEvaluated).length * numberIter > 10000 ) {
        goodScore = 0; 
        return [-1];
    }        
    var renderedSco = [];
    for(t=0; t < numberIter; t++) {
        renderedSco = renderedSco.concat(eval(expressionEvaluated));
    }
    return renderedSco;
}
function preScoRepExpr() {
}
addFunctionsTypeItem("scoRepExpr", "score");
addFunctionsTypeItem("preScoRepExpr", "prov"); 

// reindexes the internal addresses according to a prefix
// if address is [0,0,1] and prefis is "0,1", reindexed address will be [0,1,0,0,1] 
function reindexInternalAddresses(expressionToIndex, prefix) {
    expressionToIndex = compressExpr(expressionToIndex);
    var replaceStr = "scoInter([" + prefix;
    expressionToIndex = expressionToIndex.replace(/scoInter\(\[/g, replaceStr); 
    replaceStr = "scoRepExpr([" + prefix;
    expressionToIndex = expressionToIndex.replace(/scoRepExpr\(\[/g, replaceStr); 
    return expressionToIndex;
}


////////////////// functions directories

var integ_functions = ["rInt", "randInt"];
var integ_primitive = ["intPrim"];
var float_functions = ["rFlo", "sqr", "sum", "dif", "absDif", "prod", "ratio", "mod", "invers", "root", "pyth", "log", "sin", "cos", "asin", "acos", "tan", "atan", "abs", "randFlo"];
var float_primitive = ["floPrim"];
var value_functions = [];
value_functions = value_functions.concat(integ_functions, float_functions);
var value_primitive = [];
value_primitive = value_primitive.concat(integ_primitive, float_primitive);
var score_functions = ["scoTransp", "scoRep", "scoConcat", "scoRand", "scoRandFlo", "scoEnumInt", "scoMutate", "scoRender", "scoRetrog", "scoInvert", "scoExpand", "scoExcerpt", "scoPermut", "scoPermutIter", "scoSteps", "scoExpandIter", "scoMod", "scoTranspMatrix", "scoPitchClass", "scoPCGen", "scoSin", "scoCos", "scoExcerptMulti", "scoExcerptMulti2", "scoExcerptMultiTransp", "addBABBAGEvoice", "scoJoin2Scores", "scoJoin3Scores", "scoJoin4Scores", "scoJoin5Scores", "scoBABBAGE5voicesHarmonicGlobal", "scoBABBAGEharmonicVoice", "scoBABBAGEharmonicVoiceGlobal"];
var score_primitive = ["scoRandPrim", "scoPCGenPrim", "preScoInter", "preScoRepExpr", "scoNameADA", "scoNameADAintervals", "scoNameBABBAGE", "scoNameBABBAGEintervals"];
var insco_functions = ["scoInterSco"];
var insco_primitive = []; // each expression includes here the tree address of internal subscores
var pcset_functions = ["scoPCGen"];
var pcset_primitive = ["scoPCGenPrim"];

var recur_functions = ["preEncRecursio", "preEncRecursioBABBAGE"];
var recur_primitive = ["preEncRecursio", "preEncRecursioBABBAGE"];


var all_reducible_functions = ["intPrim", "floPrim", "scoRandPrim", "scoPCGenPrim"];
var reducible_score_functions = ["scoRandPrim", "scoPCGenPrim"]; // function with output array, for adding [ ] to array

var all_evolvFunctions = ["scoConcat", "scoTranspMatrix", "scoMutate", "scoMod", "scoTransp", "scoRep", "scoPermut", "scoPermutIter", "scoExpand", "scoExpandIter", "scoExcerpt", "scoRetrog", "scoInvert", "addBABBAGEvoice"]; // functions for evolution of an expression
var grow_evolvFunctions = ["scoConcat", "scoTranspMatrix", "scoExpandIter", "scoPermutIter", "scoRep", "scoExcerptMultiTransp", "addBABBAGEvoice"];
var transform_evolvFunctions = ["scoMutate", "scoMod", "scoTransp", "scoPermut", "scoExpand", "scoExcerpt", "scoRetrog", "scoInvert", "scoPitchClass", "scoExcerptMulti", "scoExcerptMulti2", "scoExcerptMultiTransp"];

var available_functions = []; // not including primitive functions
available_functions = available_functions.concat(value_functions, score_functions);



var tests_functions = ["scoRecursio", "scoBABBAGE", "scoADA", "scoADA2voices", "scoADA3voices", "scoADA4voices", "scoBABBAGE5voicesHarmonic", "scoBABBAGE5voicesHarmonicGlobal"];//, "scoExcerptMultiTransp", "addBABBAGEvoice"];//"scoExcerptMulti", "scoExcerptMulti2"];//"scoRecursioBABBAGE",  ];


// list available functions
function getFunctionsLibrary() {
    available_functions.sort(); // reorder alphabet.
    post("library of available functions:\n--------------------------\n");
    for (i=0; i<available_functions.length; i++) {
        post(available_functions[i] + " (" + functionsType[available_functions[i]] + ") - index " + function2index[available_functions[i]] + "\n");
    }
}

getFunctionsLibrary();

/////// auxiliary functions

// encode an expression as an array


// convert GenoMus native pitch scale to MIDI scale: 8.0 (GenoMus) = 60 (MIDI), 9.0 (GenoMus) = 72 (MIDI)
function genomusPitch2MIDI (genomusPitchScore) {
    midiScore = [];
    for (a=0; a<genomusPitchScore.length; a++) {
        midiScore.push(genomusPitchScore[a]*120);
    }    
    return midiScore;
}
addFunctionsTypeItem("genomusPitch2MIDI", "score");

// convert GenoMus native pitch scale to MIDI scale: 8.0 (GenoMus) = 60 (MIDI), 9.0 (GenoMus) = 72 (MIDI)
function MIDI2genomusPitch (MIDIScore) {
    genomusPitch = [];
    for (a=0; a<MIDIScore.length; a++) {
        genomusPitch.push( MIDIScore[a]/120 );
    }
    return genomusPitch;
}
addFunctionsTypeItem("MIDI2genomusPitch", "score");

// extract a pitch class from a score (respecting chromatic content of notes, and avoiding repetitions
function pcsFromScore(score_array) {
    // number of elements of PCS from 1 to chromaticQuantiz
    var numberPitchClasses = (score_array.length % chromaticQuantiz) + 1;
    var pitchClassSet = [];
    var newElem;
    var iter = 0;
    do {
        newElem = Math.round(( score_array[iter] % 1 ) * chromaticQuantiz);
        if ( pitchClassSet.indexOf(newElem) == -1 ) {
            pitchClassSet.push(newElem);
        }
        iter++;
    } while ( pitchClassSet.length < numberPitchClasses && iter < score_array.length );
    return pitchClassSet.sort(function(a,b){return a-b});
}
addFunctionsTypeItem("pcsFromScore", "score");

// return the sum of absolute values of arr
function absSumArray(arr) {
    var total = 0;
    for(a=0; a<arr.length; a++) {
        total = total + Math.abs(arr[a]);
    }
    return total;
}
addFunctionsTypeItem("absSumArray", "score");


/////////////////////
// functions to change constants
function setMaxParenthLevel(x) {
    maxLevels = x;
}  
function setMaxScoreSize(x) {
    maxScoreSize = x;
}  
function setMinScoreSize(x) {
    minScoreSize = x;
} 
function setMaxSearchIterarions(x) {
    maxSearchIterarions = x;
}  
function setReadablePrinting(x) {
    readablePrinting = x;
    printExpression(currentExpression); 
}

function setVerbosity(x) {
    if ( x != 0 ) { verboseOutput = 1 }
    else { verboseOutput = 0 }
}

//// randomSeed generator (float (0, 1) ) , to do repeatable evaluations of functions

function setDispersionAlea(x) {
    dispersionAlea = x;
}  

function getDispersionAlea() {
    post("new dispersionAlea: " + dispersionAlea + "\n");
}

var initSeed = 0.93459283; // seed ( 0 < seed < 1 )
var seed = initSeed; // this value change every time randomSeed function is called
function setInitSeed() {
    initSeed = Math.random();
    seed = initSeed;
}

function setManualSeed(x) {
    initSeed = x;
    seed = initSeed;
}

function getSeed() {
    post("new initSeed: " + initSeed + "\n");
}

function randomSeed() {
    seed = (seed*dispersionAlea*(1-seed)) % 1;
    return seed;
}  


/////////////////////
// primitive funcions for expression manipulations

// add a right parenthesis to expression
function addRightParenth(expression) { 
    openParenth--;
    return expression + ")"; 
}

// add a comma
function addComma(expression) { 
    argumentStatus[openParenth]--;
    return expression + ", "; 
}

// return an integer constant from an interval
function rInt(value_1, value_2) {
    var minimum = Math.min(value_1, value_2);
    var maximum = Math.max(value_1, value_2) + 1;
    return Math.floor(randomSeed() * (maximum-minimum) + minimum);
}

// return an integer constant 
function intPrim() {
    return rInt(0, 127);
}

// return a float constant from an interval
function rFlo(value_1, value_2) {
    var minimum = Math.min(value_1, value_2);
    var maximum = Math.max(value_1, value_2); 
    return randomSeed() * (maximum-minimum) + minimum; 
}

// return a float constant 
function floPrim() {
    return rFlo(0, 127);
}

// add an integer constant to expression
function addIntConstant(expression, limit1, limit2) { 
    return expression + rInt(limit1, limit2); 
}

// add a float constant to expression
function addFloatConstant(expression, limit1, limit2) { 
    return expression + rFlo(limit1, limit2); 
}

// add a functionName to expression
function addFunctionName(expression, functionType, primitiveFlag) { 
    if (primitiveFlag == 0) {  
        var functions_set = eval(functionType + "_functions");
    }
    else {
        var functions_set = eval(functionType + "_primitive");
    }    
    openParenth++;
    currentFunctionName = functions_set[rInt(0, functions_set.length - 1)];
    usedFunctionsArray[openParenth - 1] = currentFunctionName;
    var currentArgument = getArgs(eval(String(usedFunctionsArray[openParenth-1]))).length;
    argumentStatus[openParenth] = eval(currentFunctionName).length - 1;  
    return expression + currentFunctionName + "(";    
}

 


// function to get the arguments type of a function (to determine how call to new functions)
// get all arguments name as array
function getArgs(func) {
    var str = func.toString().match(/\([\w*\s*\,*]*\)/).toString();
    return  str.substring(1, str.length - 1).split(",");
}

// get the first five characters of an argument (to compare and delete index of several arguments and retain only type)
function getArgType(functionName, argumentNum) {
    if (argumentNum == 0) {
        return getArgs(functionName)[argumentNum].slice(0,5);
    }
    else
    {
        return getArgs(functionName)[argumentNum].slice(1,6);
    }
}   

// compress an expanded expression
function compressExpr(expandedFormExpr) {
    var temporaryExpr = "";	
    for ( charIndx = 0; charIndx < expandedFormExpr.length; charIndx++) {
        if ( expandedFormExpr.charAt(charIndx) != " " && expandedFormExpr.charAt(charIndx) != "\n") {
            temporaryExpr = temporaryExpr + expandedFormExpr.charAt(charIndx);
        }
    }
    temporaryExpr = temporaryExpr.replace(/,/g,", ");
    expandedFormExpr = temporaryExpr;
    return expandedFormExpr;
}

// expand a compressed expression in a readable form
function expandExpr(compressedFormExpr) {
    compressExpr(compressedFormExpr);	
    var expandedExpression = "";	
    compressedFormExpr = compressedFormExpr.replace(/\(/g,"(\n");
    compressedFormExpr = compressedFormExpr.replace(/, /g,",\n");
    compressedFormExpr = compressedFormExpr.replace(/\n\)/g,")");
    var parenthCount = 0;
    for ( charIndx = 0; charIndx < compressedFormExpr.length; charIndx++) {
        expandedExpression = expandedExpression + compressedFormExpr.charAt(charIndx);
        if ( compressedFormExpr.charAt(charIndx) == "(" ) { parenthCount++ }
        if ( compressedFormExpr.charAt(charIndx) == ")" ) { parenthCount-- }
        if ( compressedFormExpr.charAt(charIndx) == "\n" ) { 
            var tabulation = "    ";
            for ( n = 0; n < parenthCount; n++ ) {
                expandedExpression = expandedExpression + tabulation;
            }
        }
    }
    // rewrite expandedExpr mantaining matrices in a single line
    var matrixCompactExpr = "";
    var matrixOpen = 0;
    for ( charIndx = 0; charIndx < expandedExpression.length; charIndx++) {
        if ( expandedExpression.charAt(charIndx) == "[" ) { matrixOpen++ };
        if ( expandedExpression.charAt(charIndx) == "]" ) { matrixOpen-- };
        if ( matrixOpen > 0 ) {
            if ( expandedExpression.charAt(charIndx) != "\n" && expandedExpression.charAt(charIndx) != " " ) {
            matrixCompactExpr = matrixCompactExpr + expandedExpression.charAt(charIndx);
            }
        }
        else {
            matrixCompactExpr = matrixCompactExpr + expandedExpression.charAt(charIndx);
        }
    }         
    compressedFormExpr = matrixCompactExpr; 
    return compressedFormExpr;;
}

// print currentExpression
function printExpression(expr) { 
    if ( readablePrinting == 1 ) { 
        outlet( 1, expandExpr(expr) );
        if ( verboseOutput == 1 ) {
            post( compressExpr(expr) + "\n" );	
        }
    }	   	
    else { 
        outlet( 1, compressExpr(expr) );
        if ( verboseOutput == 1 ) {
            post( expr + "\n" );	
        }
    }	
} 

function writeScore(newWrittenExpression ) {
    seed = initSeed; 
    var generatedScore = ""; 
    generatedScore = eval(newWrittenExpression);
    generatedScoreLength = generatedScore.length
    if (goodScore == 1 && generatedScoreLength >= minScoreSize && generatedScoreLength <= maxScoreSize ) {
        if ( isBranch == 0 && verboseOutput == 1 ) { 
            post("generated score: " + generatedScore + "\n");
            post("score length: " + generatedScoreLength  + "\n"); 
        }    
        outlet(0, "clear"); 
        outlet(2, generatedScore);
        for(i=0; i< generatedScoreLength; i++) {
            outlet(0, generatedScore[i]);
            outlet(0, "cr");
        }   
        if ( isBranch == 0 ) { printExpression( newWrittenExpression ) };
        goodScoreSize = 1;
    }  
}  
    
// look for a variation of the same expression with different seed
function rewriteScore(newWrittenExpression) {
    goodScoreSize = 0;  
    numIter = 0;
    var newGeneratedScore = "";
    do { 
        numIter++;  
        setInitSeed();
        newGeneratedScore = eval( newWrittenExpression );
    } while ( (newGeneratedScore.length < minScoreSize || newGeneratedScore.length > maxScoreSize ) && numIter < maxSearchIterarions ); 
    if ( verboseOutput == 1 ) {
        getSeed();
        post("generated score: " + newGeneratedScore + "\n");
        post("score length: " + newGeneratedScore.length  + "\n"); 
    }
    outlet(0, "clear"); 
    outlet(2, newGeneratedScore);
    for(i=0; i< newGeneratedScore.length; i++) {
        outlet(0, newGeneratedScore[i]);
        outlet(0, "cr");
    }
    printExpression( newWrittenExpression );
    goodScoreSize = 1;
    if ( verboseOutput == 1 ) {
        post("rewriting iterations: " + numIter + "\n");
    }
}  

//////////// CORE FUNCTION
// function to create functions from scratch from library of functions
function createExpression(expressionType) {
    var createExpressionIteration = 0;
    var internalExpression, numFunctions;
    newExpression = "";
    do {
        createExpressionIteration++;  
        // init goodScore  
        goodScore = 1; 
        goodScoreSize = 0;  
        // init new function
        var createdExpression = "";
        usedFunctionsArray = [];
        var activeArgument;
        var functionTypeAsked;
        openParenth = 0;
        var writingAutomataStatus = 0;
        // write the first functionName
        createdExpression = addFunctionName(createdExpression,expressionType,0);
        // recursive writing process
        numFunctions = 0;
                while (openParenth > 0) {
            // after writing a functionName
            switch (writingAutomataStatus) {
                // last written: functionName or comma
                case 0:
                    activeArgument =  (getArgs(eval(String(usedFunctionsArray[openParenth-1]))).length - argumentStatus[openParenth] -1) ;
                    functionTypeAsked = getArgType(eval(String(usedFunctionsArray[openParenth-1])), activeArgument);2
                    if (Math.random() < .5 && openParenth < maxLevels) {  
                        createdExpression = addFunctionName(createdExpression, functionTypeAsked, 0);  
                        numFunctions++;
                        // number of functions used is limited to constant maxNumberFuncions, to avoid too big expressions
                        if ( numFunctions > maxNumberFuncions ) { 
                            //post("error: too many functions\n");
                            break;
                        }
                    }
                    else {
                        createdExpression = addFunctionName(createdExpression, functionTypeAsked, 1);
                        writingAutomataStatus = 1;     
                        numFunctions++;
                    }
                    break;
                // last written: constant or right parenthesis
                case 1:  
                    if ( argumentStatus[openParenth] > 0 ) {
                        createdExpression = addComma(createdExpression);
                        writingAutomataStatus = 0;
                    }
                    else {
                        createdExpression = addRightParenth(createdExpression);
                        writingAutomataStatus = 1;
                    }
                    break;   
            }
        }
        // writes autoreferences, if needed
        if ( createdExpression.search("preScoInter()") != -1 ) {
            internalExpression = "scoInter([0," + eval(insco_primitive[rInt(0, insco_primitive.length - 1)]) + "])";
            createdExpression = createdExpression.replace(/preScoInter\(\)/g, internalExpression);
        }
        if ( createdExpression.search("preScoRepExpr()") != -1 ) {
            internalExpression = "scoRepExpr([0," + eval(insco_primitive[rInt(0, insco_primitive.length - 1)]) + "], intPrim())";
            createdExpression = createdExpression.replace(/preScoRepExpr\(\)/g, internalExpression);
        }
        if ( createdExpression.search("preEncRecursio()") != -1 ) {
            internalExpression = "[" + encodeExpr( createRecursio(4) ) + "]";
            createdExpression = createdExpression.replace(/preEncRecursio\(\)/g, internalExpression);
        }
        if ( createdExpression.search("preEncRecursioBABBAGE()") != -1 ) {
            internalExpression = "[" + encodeExpr( createRecursio(6) ) + "]";
            createdExpression = createdExpression.replace(/preEncRecursioBABBAGE\(\)/g, internalExpression);
        }
        newExpression = createdExpression;
        writeScore( newExpression );    
        if ( isBranch == 1 ) { goodScoreSize = 1 };
        // post("Branch=" + isBranch + " goodSco=" + goodScore + " goodScoreSize=" + goodScoreSize + "\n");
    } while ( createExpressionIteration < maxSearchIterarions && (goodScore == 0 || goodScoreSize == 0) );
    if ( isBranch == 0 ) { post("expr. creation iterations: " + createExpressionIteration + "\n") };
    return newExpression;
}  

//// automatic evolution of currentExpression, using only functions of functionSet as the basis function of the evolved new expression
function evolveCurrentExpression(functionSet) {
    var evolIter = 0;
    var newRightBranch, evolvMainFunction, evolvMainFunctionArgType, numberNewBranches, memoMinScoreSize;
    provisoryExpression = reindexInternalAddresses(currentExpression, "0,");
    isBranch = 1;
    do {
        newRightBranch = [];
        evolvMainFunction = eval(functionSet)[rInt(0, eval(functionSet).length - 1)]; // chooses a function from the evolv_functions set
        numberNewBranches = getArgs(eval(evolvMainFunction)).length - 1; // look for the number of new branches required
        growingExpression = evolvMainFunction + "(" + provisoryExpression; // + ", " + ")"; // compose a provisional but correct expression to avoid syntax errors
        for (newBranch = 1; newBranch <= numberNewBranches ; newBranch++) {
            evolvMainFunctionArgType = getArgType(eval(evolvMainFunction), newBranch);
            newRightBranch = createExpression( evolvMainFunctionArgType );
            // control if autoreference in new branchs is compulsory
            if ( alwaysAutoreference == 1 && newRightBranch.search("scoInter") == -1 && newRightBranch.search("scoRepExpr") == -1 ) { 
                goodScore = 0; 
            }
            growingExpression = growingExpression + ", " + newRightBranch;
        }
        growingExpression = growingExpression + ")";
        memoMinScoreSize = minScoreSize;  // suspends the evaluation of scoreLength while constructing new branchs      
        maxScoreSize = maxScoreSize * 10;
        writeScore( growingExpression );	
        maxScoreSize = maxScoreSize / 10;
        minScoreSize = memoMinScoreSize;
        evolIter++;
        if ( generatedScoreLength < minScoreSize || generatedScoreLength > maxScoreSize ) { goodScoreSize = 0 }; 
    } while ( (goodScore != 1 || goodScoreSize != 1) && evolIter < 20 );	 // allow only 20 iterations for evolv
    if ( alwaysAutoreference == 1 ) { alwaysAutoreference = 0 }; 
    if ( goodScoreSize == 1 ) {
        printExpression( growingExpression )
        if ( verboseOutput == 1 ) {
            post("score length: " + generatedScoreLength  + "\n"); 
        }
        post("evolv. iterations: " + evolIter + "\n"); 
    }
    else {
        post("solution not found" + "\n");
    }
    separator();
}  

// types of evolution 
function evolvExpr() { 
    evolveCurrentExpression("all_evolvFunctions");
}

function growExpr() { 
    alwaysAutoreference = 1;
    evolveCurrentExpression("grow_evolvFunctions");
}

function transformExpr() { 
    evolveCurrentExpression("transform_evolvFunctions");
}

function testExpr() { 
    evolveCurrentExpression("tests_evolvFunctions");
}

// convert primitives in numeric values
function renderPrimitiveFunctions( expressionToConvert ) {
    var regExpress, replaceExpr;
    for (priFunctNum = 0; priFunctNum < reducible_score_functions.length; priFunctNum++) {
            functionToReduce = reducible_score_functions[priFunctNum];
            regExpress = "/" + functionToReduce + "\\(\\)/g";
            putKlammers = "[" + functionToReduce + "()]"; 
            expressionToConvert = expressionToConvert.replace( eval(regExpress), putKlammers );
    }
    for (priFunctNum2 = 0; priFunctNum2 < all_reducible_functions.length; priFunctNum2++) {
        do {            
            functionToReduce = all_reducible_functions[priFunctNum2];
            regExpress = "/" + functionToReduce + "\\(\\)/";
            replaceExpr = eval(functionToReduce);
            expressionToConvert = expressionToConvert.replace( eval(regExpress), replaceExpr);
        } while ( expressionToConvert.search(functionToReduce) != -1 );
    }
    return expressionToConvert;
}


/*
function testRecursioConv(x) {
    updateCurrentExpression();
    recursioTestExp = createRecursiveFunction( currentExpression, x );
    printExpression( recursioTestExp );
} 
*/

function renderAllPrimitives() {
    updateCurrentExpression();
    currentExpressionRendered = renderPrimitiveFunctions( currentExpression );
    rewriteScore( currentExpressionRendered );
}

function unrenderAllPrimitives() {
    text (currentExpression);
}

function updateRenderedCurrentExpression() {
    text (currentExpressionRendered);
}

////// activate functions from Max interface

function createNewExpression() {
    isBranch = 0;
    insco_primitive = [];
    currentExpression = createExpression("available"); 
    growingExpression = currentExpression;
    mapSubScores( currentExpression );
    separator();
}

function createTestExpression() {
    isBranch = 0;
    insco_primitive = [];
    currentExpression = createExpression("tests"); 
    growingExpression = currentExpression;
    mapSubScores( currentExpression );
    separator();
}

function writeSco() {
    writeScore( currentExpression );
    separator();
}

function rewriteSco() {
    rewriteScore( currentExpression );
    separator();
}

function separator() {
    if ( verboseOutput == 1 ) {
        post("--------------------\n");
    }   
}

////////////////////////// functions to manipulate, store and recall current expressions   
   
// introduce manualmente una expression desde fuera del script (utiliza la palabra text que preceden los mensajes del modulo de Max
function text(externalExp) {
    //currentExpression, 
    growingExpression = compressExpr(externalExp);
    updateCurrentExpression();
    seed = initSeed;  
    var generatScore = eval(currentExpression);
    if ( verboseOutput == 1 ) {
        post("activating external expression: " + currentExpression + "\n");
        post("generatedScore = " + generatScore + "\n");
        post("score length = " + generatScore.length  + "\n");
    } 
    outlet(0, "clear"); 
    outlet(2, generatScore);
    // send score to text object in Max interface
    for(i=0; i< generatScore.length; i++) {
        outlet(0, generatScore[i]);
        outlet(0, "cr");
    }
    printExpression( currentExpression );
}

// salva la expresion en un array de cadenas y devuelve la expresion por la salida 2
function saveExpression() {
    savedExpressions[savedExpressions.length] = currentExpression;
    post("total saved expressions: " + savedExpressions.length + "\n");
}

// lista las expresiones guardadas
function listExpressions() {
    for (i=0; i<savedExpressions.length; i++)
    {
        post("expression " + i + ": " + savedExpressions[i] + "\n");
    }
}

// lee una de las expresiones guardadas para ser activada
function loadExpression(expression_number) {
    currentExpression = savedExpressions[expression_number];
    printExpression( currentExpression );
}

function clearExpressionsList() {
    savedExpressions = [];
    post("total saved expressions: " + savedExpressions.length + "\n");
}

////////////// functions to evolve an expressions

// functions to repeat the current expression
function repeatCurrentExpression() {
    currentExpression = reindexInternalAddresses(currentExpression, "0,");
    growingExpression = "scoRep(" + currentExpression + ", " + "intPrim())";
    updateCurrentExpression();
}
		
// uses the approved growingExpression as the new updated currentExpression
function updateCurrentExpression() {
    currentExpression = growingExpression;
    printExpression( currentExpression );
    mapSubScores( currentExpression );
    outlet(3, generatedScoreLength);
}


// convert a expression to a score, to use like data into another expression
function currentScore2expression() {
    seed = initSeed;  
    var convertedScore = [];
    convertedScore = "scoRender([" + (eval(growingExpression)).join(", ") + "])";
    growingExpression = convertedScore;
    updateCurrentExpression();
}


// compare 2 arrays (auxiliary function for subExpr), returns 1 (0) if both arrays are equal (diff.), 
function compareArrays(addr1, addr2) {
    var equality = true;
    var positionEvalued = 0;
    do {
        if ( addr1[positionEvalued] != addr2[positionEvalued] ) { 
            equality = false;
        }	
        positionEvalued++;
    } while ( positionEvalued < Math.max(addr1.length, addr2.length) && equality == true );
    return equality;
}
	
// extract a subExpression corresponding to the address position (string tree form) from expression
function extractSubExpression(expression, address) {
    address = eval(address); // converts address from string to array
    var addressPosition = [0]; // momentan address in expression
    var level = 0; // matrix position (level of the branch)
    var matchAddresses = false; // boolean to compare addresses
    // look for the position of the address
    for ( charIndx = 0; charIndx < expression.length && compareArrays(address, addressPosition) == false; charIndx++) {
        if (expression.charAt(charIndx) == "(" && expression.charAt(charIndx+1) != ")" ) { 
            level++;
            if ( addressPosition[level] == undefined ) { addressPosition[level] = 0 }
            else { 
                addressPosition[level] = addressPosition[level] + 1;
            }
        }
        if (expression.charAt(charIndx) == "," ) { 
            addressPosition[level] = addressPosition[level] + 1;
        }
        if (expression.charAt(charIndx) == ")" && expression.charAt(charIndx-1) != "(" ) { 
            addressPosition.pop();
            level--;
        }
    }
    // copy and return the selected substring 
    if ( charIndx == expression.length ) { post("error: substring not found") }
    else {
        var subExpr = "";
        level = 0;
        copyChar = charIndx;
        do {
            subExpr = subExpr + expression.charAt(copyChar);
            if ( expression.charAt(copyChar) == "(" ) { level++ }; 
            copyChar++;
        } while ( level == 0 && copyChar <= expression.length );
        do {
            subExpr = subExpr + expression.charAt(copyChar);
            if ( expression.charAt(copyChar) == "(" ) { level++ } ;
            if ( expression.charAt(copyChar) == ")" ) { level-- } ;
            copyChar++;			
        } while ( level != 0 && copyChar <= expression.length );
        return subExpr;
    }
}

// return a map of the subScores from a given expression,
// and store it in an array used like variables for certain functions
function mapSubScores(expression) {
    insco_primitive = [];
    var addressPosit = [0]; // momentan address in expression
    var levl = 0; // matrix position (level of the branch)
    var numSubSco = 0;
    // look for the position of the address
    for ( charIndx = 0; charIndx < expression.length; charIndx++) {
        if (expression.charAt(charIndx) == "(" && expression.charAt(charIndx+1) != ")" ) { 
            levl++;
            if ( addressPosit[levl] == undefined ) { addressPosit[levl] = 0 }
            else { 
                addressPosit[levl] = addressPosit[levl] + 1;
            }
        }	
        if (expression.charAt(charIndx) == "," ) { 
            addressPosit[levl] = addressPosit[levl] + 1;
        }	
        if (expression.charAt(charIndx) == ")" && expression.charAt(charIndx-1) != "(" ) { 
            addressPosit.pop();
            levl--;
        }
        if (expression.substr(charIndx,3) == "sco") {
            insco_primitive[numSubSco] = "[" + addressPosit.toString() + "]";
            numSubSco++;
        }	
    }
    if ( verboseOutput == 1 ) {
        post("internal scores addresses: " + insco_primitive + "\n");
    }
}

function createRecursiveFunction( expressionToConvert, numRecursiveVariables ) {
    var regExpress, replaceExpr, randVal;
    for (priFunctNum = 0; priFunctNum < reducible_score_functions.length; priFunctNum++) {
            functionToReduce = reducible_score_functions[priFunctNum];
            regExpress = "/" + functionToReduce + "\\(\\)/g";
            putKlammers = "[" + functionToReduce + "()]"; 
            expressionToConvert = expressionToConvert.replace( eval(regExpress), putKlammers );
    }
    for (priFunctNum2 = 0; priFunctNum2 < all_reducible_functions.length; priFunctNum2++) {
        do {            
            functionToReduce = all_reducible_functions[priFunctNum2];
            regExpress = "/" + functionToReduce + "\\(\\)/";
            // substitutions types (using last intervals, last notes, or constants)
            var randVal = 0;//randomSeed();
            if ( randVal <= randomSeed() ) {
                replaceExpr = "intrv(s, " + rInt(1, numRecursiveVariables) + ")";
            }              
            else {
                replaceExpr = eval(functionToReduce);
            }
            expressionToConvert = expressionToConvert.replace( eval(regExpress), replaceExpr);
        } while ( expressionToConvert.search(functionToReduce) != -1 );
    }
    return expressionToConvert;
}

function createRecursio(intervalsIntoAccount) {
    isBranch = 1;
    //var retainMaxNotes = maxScoreSize;
    //maxScoreSize = 50000;
    var recursiveExpression = createRecursiveFunction( createExpression("value"), intervalsIntoAccount);
    post(recursiveExpression + "\n");
    //maxScoreSize = retainMaxNotes;    
    return recursiveExpression;
}


//////// encoding / decoding expressions

function encodeExpr (expr) {
    // arreglar el bug cuando encuentra una cadena de letras que no termina en "(" y entra en loop infinito
    // first compress expr
    expr = compressExpr(expr);
    var encodedExpr = [];
    var asciiVal, subStrRead, arrayRead;
    for( charIndex = 0; charIndex < expr.length; charIndex++ ) {
        // if found a function, writes "1" followed by the index function name
        // first, function name is read, from the first character to "(", or a non name character ir found
        asciiVal = expr.charCodeAt(charIndex);
        if ( (asciiVal > 64 && asciiVal < 91) || (asciiVal > 96 && asciiVal < 123) ) {
            subStrRead = "";
            while ( (asciiVal > 64 && asciiVal < 91) || (asciiVal > 94 && asciiVal < 123) || (asciiVal > 47 && asciiVal < 58) ) {
                subStrRead = subStrRead + expr[charIndex];
                charIndex++;
                asciiVal = expr.charCodeAt(charIndex);
            }
            charIndex--;
            //charIndex--;
            // if function name is indexed, flag is 1 followed by the function index
            // else, flag is -1, followed by the length of this substring, and the substring itself coded as ascii numbers 
            if ( (function2index[subStrRead]) == undefined ) {
                encodedExpr.push(-1); // 1 is the flag for a non indexed string
                encodedExpr.push(subStrRead.length); // 1 is the flag for a non indexed string
                // 1 is the flag for a non indexed string
                for ( strch = 0; strch < subStrRead.length; strch++) {
                    encodedExpr.push(subStrRead.charCodeAt(strch));
                }
            }
            else {
                encodedExpr.push(1); // 1 is the flag for a function name
                encodedExpr.push(function2index[subStrRead]);
            }
        }
        // if "(" is found, adds "5"
        else if ( asciiVal == 40 ) {
            encodedExpr.push(5); // 0 is the flag for a left parenth
        }
        // if ")" is found, adds "0"
        else if ( asciiVal == 41 ) {
            encodedExpr.push(0); // 0 is the flag for a right parenth
        }
        // if a single comma is found (not as part of an array), adds flag "2"
        else if ( asciiVal == 44 ) {
            encodedExpr.push(2);
        }
        // if a single number is found, is preceded by flag 3
        else if ( (asciiVal > 44 && asciiVal < 47) || (asciiVal > 47 && asciiVal < 58) ) {
            subStrRead = "";
            // read number as string
            while ( (asciiVal > 44 && asciiVal < 47) || (asciiVal > 47 && asciiVal < 58) ) {
                asciiVal = expr.charCodeAt(charIndex);
                subStrRead = subStrRead + expr[charIndex];   
                charIndex++;                                   
            }
            // convert substring into actual number
            encodedExpr.push(3);
            encodedExpr.push(parseFloat(subStrRead));
            charIndex--;
            charIndex--;
        }    
        // if "[" is found, adds the flag 4, the lenght of the array, and parses the expression as actual numbers into de encoded array, until "]" is found
        // example: "[3, 45.2, 0, 1.03]" would be encoded as 3, 4, 3, 45.2, 0, 1.03
        else if ( asciiVal == 91 ) {
            encodedExpr.push(4); // 4 is the flag for "["
            charIndex++;                                              
            arrayRead = [];
            do {
                subStrRead = "";
                // read numbers as strings
                do {
                    asciiVal = expr.charCodeAt(charIndex);
                    if ( (asciiVal > 44 && asciiVal < 47) || (asciiVal > 47 && asciiVal < 58) ) {
                        subStrRead = subStrRead + expr[charIndex];   
                    }
                    charIndex++;                                   
                } while ( expr[charIndex] != "," && expr[charIndex] != "]" );
                // convert substring into actual number
                arrayRead.push(parseFloat(subStrRead));
            } while ( expr[charIndex] != "]" );
            encodedExpr.push(arrayRead.length); // arrayRead.length tells how long is the numeric array
            encodedExpr = encodedExpr.concat(arrayRead);
        }
    }
    return encodedExpr;
}
addFunctionsTypeItem("encodeExpr", "score"); 


function decodeExpr (score) {
    var arrayLength;
    var decodedExpr = "";    
    for ( scoElem = 0; scoElem < score.length; scoElem++ ) {
        switch (score[scoElem]) {
            // right parenth
            case 0:
                decodedExpr = decodedExpr + ")";
                break;
            // function name
            case 1:
                scoElem++;
                decodedExpr = decodedExpr + index2function[score[scoElem]];
                break;
            // comma
            case 2:
                decodedExpr = decodedExpr + ", ";
                break;
            // single value
            case 3:
                scoElem++;
                decodedExpr = decodedExpr + score[scoElem];
                break;
            // numeric array
            case 4:
                scoElem++;
                decodedExpr = decodedExpr + "[";
                arrayLength = score[scoElem];
                scoElem++;
                for( arrEl=0; arrEl < arrayLength - 1; arrEl++ ) {
                    decodedExpr = decodedExpr + score[scoElem] + ", ";
                    scoElem++;
                }
                decodedExpr = decodedExpr + score[scoElem] + "]";
                break;
            // non indexed substring
            case 5:
            // left parenth
                decodedExpr = decodedExpr + "(";
                break;
            case -1:
                scoElem++;
                arrayLength = score[scoElem];
                for( el=0; el < arrayLength; el++ ) {
                    scoElem++;
                    decodedExpr = decodedExpr + String.fromCharCode(score[scoElem])
                }
                break;
            // bad expression  
            default:
                post("error: probably bad encoded expression" + "\n");
                goodScore = 0;
                return [-1];
        }
    }
    return decodedExpr;
}
addFunctionsTypeItem("decodeExpr", "expre"); 
